// Perform write operations for sshsync

#define _GNU_SOURCE

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "proto.h"

#define MIN(x, y) ({ \
	typeof(x) __x = (x); \
	typeof(y) __y = (y); \
	__x < __y ? __x : __y; \
})

extern void assert_error(void) __attribute__((error("bad assertion")));

static void assert_fflush(FILE *stream);
static char *get_string(void);
static void process_args(int argc);
static void put_file(void);

int main(int argc, char **argv) {
	process_args(argc);

	if (chdir(argv[1]))
		err(EXIT_FAILURE, "chdir");

	int command;
	for (command = getchar(); command != EOF; command = getchar()) {
		switch (command) {
		case PUT:
			put_file();
			break;
		case SYNC:
			if (putchar('\0') == EOF)
				err(EXIT_FAILURE, "putchar");
			assert_fflush(stdout);
			break;
		default:
			errx(EXIT_FAILURE, "Unsupported command");
		}
	}

	return EXIT_SUCCESS;
}

static void assert_fflush(FILE *stream) {
	if (fflush(stream))
		err(EXIT_FAILURE, "fflush");
}

static int assert_getchar(void) {
	int c = getchar();
	if (c == EOF)
		err(EXIT_FAILURE, "getchar");
	return c;
}

static void *assert_realloc(void *ptr, size_t size) {
	void *new_ptr = realloc(ptr, size);
	if (!new_ptr)
		err(EXIT_FAILURE, "realloc");
	return new_ptr;
}

static int32_t get_int32(void) {
	int32_t i = 0;
	int shift;
	for (shift = 0; shift < 32; shift += 8)
		i |= assert_getchar() << shift;
	return i;
}

static mode_t get_mode(void) {
	mode_t mode = assert_getchar();
	mode |= assert_getchar() << 8;
	return mode;
}

static size_t get_size(void) {
	size_t size = 0;
	int shift = 0;
	int c;
	do {
		c = assert_getchar();
		_Bool too_big = 0;
		switch (sizeof(size_t)) {
		case 4:
			too_big = shift == 28 && c & 0xf0;
			break;
		case 8:
			too_big = shift == 63 && c & 0xfe;
			break;
		default:
			assert_error();
		}
		if (too_big)
			errx(EXIT_FAILURE, "Size too big");
		size |= (size_t)(c & 0x7f) << shift;
		shift += 7;
	} while (c & 0x80);
	return size;
}

static char *get_string(void) {
	size_t size = 128;
	char *s = malloc(size);
	if (!s)
		errx(EXIT_FAILURE, "Memory error");
	size_t len = 0;
	for (;;) {
		int c = assert_getchar();
		s[len++] = c;
		if (!c) {
			s = assert_realloc(s, len);
			return s;
		}
		if (len == size) {
			size <<= 1;
			s = assert_realloc(s, size);
		}
	}
}

static void process_args(int argc) {
	if (argc == 2)
		return;
	errx(EXIT_FAILURE, "This program is an internal tool.");
}

static void put_file(void) {
	char *pathname = get_string();
	mode_t mode = get_mode();
	errno = 0;
	if (unlink(pathname) && errno != ENOENT)
		err(EXIT_FAILURE, NULL);
	int fd = open(pathname, O_WRONLY | O_CREAT | O_EXCL | O_NOFOLLOW,
		mode | S_IWUSR);
	if (fd == -1)
		err(EXIT_FAILURE, NULL);
	free(pathname);
	FILE *stream = fdopen(fd, "w");
	if (!stream)
		err(EXIT_FAILURE, "fdopen");

	static char buffer[BUFSIZ];
	size_t size = get_size();
	while (size) {
		size_t nmemb = MIN(size, (size_t) BUFSIZ);
		size_t nread = fread(buffer, 1, nmemb, stdin);
		if (!nread)
			err(EXIT_FAILURE, "fread");
		if (fwrite(buffer, 1, nread, stream) < nread)
			err(EXIT_FAILURE, "fwrite");
		size -= nread;
	}
	assert_fflush(stream);

	if (!(mode & S_IWUSR))
		if (fchmod(fd, mode))
			err(EXIT_FAILURE, "fchmod");
	struct timespec times[2];
	times[0].tv_nsec = UTIME_OMIT;
	times[1].tv_sec = get_int32();
	times[1].tv_nsec = get_int32();
	if (futimens(fd, times))
		err(EXIT_FAILURE, "futimens");

	if (fclose(stream))
		err(EXIT_FAILURE, "fclose");
}
